@@ -1,7 +1,7 @@
 /*
  *  Copyright (C) 2000 by Glenn McGrath
  *  Copyright (C) 2001 by Laurence Anderson
- *	
+ *
  *  Based on previous work by busybox developers and others.
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -19,13 +19,18 @@
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
+#include "config.h"
+
 #include <stdio.h>
 #include <errno.h>
 #include <stdlib.h>
+#include <stdbool.h>
 #include <string.h>
 #include <unistd.h>
 #include <utime.h>
 #include <libgen.h>
+#include <grp.h>
+#include <pwd.h>
 
 #include "libbb.h"
 
@@ -47,8 +52,8 @@ seek_by_read(FILE* fd, size_t len)
         char buf[SEEK_BUF];
 
         while (len) {
-                cc = fread(buf, sizeof(buf[0]), 
-                                len > SEEK_BUF ? SEEK_BUF : len, 
+                cc = fread(buf, sizeof(buf[0]),
+                                len > SEEK_BUF ? SEEK_BUF : len,
                                 fd);
 
                 total += cc;
@@ -74,17 +79,17 @@ seek_sub_file(FILE *fd, const int count)
 }
 
 
-/* Extract the data postioned at src_stream to either filesystem, stdout or 
- * buffer depending on the value of 'function' which is defined in libbb.h 
+/* Extract the data postioned at src_stream to either filesystem, stdout or
+ * buffer depending on the value of 'function' which is defined in libbb.h
  *
  * prefix doesnt have to be just a directory, it may prefix the filename as well.
  *
- * e.g. '/var/lib/dpkg/info/dpkg.' will extract all files to the base bath 
- * '/var/lib/dpkg/info/' and all files/dirs created in that dir will have 
+ * e.g. '/var/lib/dpkg/info/dpkg.' will extract all files to the base bath
+ * '/var/lib/dpkg/info/' and all files/dirs created in that dir will have
  * 'dpkg.' as their prefix
  *
  * For this reason if prefix does point to a dir then it must end with a
- * trailing '/' or else the last dir will be assumed to be the file prefix 
+ * trailing '/' or else the last dir will be assumed to be the file prefix
  */
 static char *
 extract_archive(FILE *src_stream, FILE *out_stream,
@@ -100,7 +105,7 @@ extract_archive(FILE *src_stream, FILE *out_stream,
 
 	*err = 0;
 
-	/* prefix doesnt have to be a proper path it may prepend 
+	/* prefix doesnt have to be a proper path it may prepend
 	 * the filename as well */
 	if (prefix != NULL) {
 		/* strip leading '/' in filename to extract as prefix may not be dir */
@@ -129,11 +134,11 @@ extract_archive(FILE *src_stream, FILE *out_stream,
 
 	if (function & extract_to_stream) {
 		if (S_ISREG(file_entry->mode)) {
-			*err = copy_file_chunk(src_stream, out_stream, file_entry->size);			
+			*err = copy_file_chunk(src_stream, out_stream, file_entry->size);
 			archive_offset += file_entry->size;
 		}
 	}
-	else if (function & extract_one_to_buffer) { 
+	else if (function & extract_one_to_buffer) {
 		if (S_ISREG(file_entry->mode)) {
 			buffer = (char *) xmalloc(file_entry->size + 1);
 			fread(buffer, 1, file_entry->size, src_stream);
@@ -189,7 +194,7 @@ extract_archive(FILE *src_stream, FILE *out_stream,
 						goto cleanup;
 					}
 					archive_offset += file_entry->size;
-					*err = copy_file_chunk(src_stream, dst_stream, file_entry->size);			
+					*err = copy_file_chunk(src_stream, dst_stream, file_entry->size);
 					fclose(dst_stream);
 				}
 				break;
@@ -230,7 +235,7 @@ extract_archive(FILE *src_stream, FILE *out_stream,
 
 		}
 
-		/* Changing a symlink's properties normally changes the properties of the 
+		/* Changing a symlink's properties normally changes the properties of the
 		 * file pointed to, so dont try and change the date or mode, lchown does
 		 * does the right thing, but isnt available in older versions of libc */
 		if (S_ISLNK(file_entry->mode)) {
@@ -247,7 +252,7 @@ extract_archive(FILE *src_stream, FILE *out_stream,
 			chmod(full_name, file_entry->mode);
 		}
 	} else {
-		/* If we arent extracting data we have to skip it, 
+		/* If we arent extracting data we have to skip it,
 		 * if data size is 0 then then just do it anyway
 		 * (saves testing for it) */
 		seek_sub_file(src_stream, file_entry->size);
@@ -256,7 +261,7 @@ extract_archive(FILE *src_stream, FILE *out_stream,
 	/* extract_list and extract_verbose_list can be used in conjunction
 	 * with one of the above four extraction functions, so do this seperately */
 	if (function & extract_verbose_list) {
-		fprintf(out_stream, "%s %d/%d %8d %s ", mode_string(file_entry->mode), 
+		fprintf(out_stream, "%s %d/%d %8d %s ", mode_string(file_entry->mode),
 			file_entry->uid, file_entry->gid,
 			(int) file_entry->size, time_string(file_entry->mtime));
 	}
@@ -295,8 +300,13 @@ unarchive(FILE *src_stream, FILE *out_stream,
 
 		if (extract_names != NULL) {
 			int found_flag = FALSE;
+			char *p = file_entry->name;
+
+			if (p[0] == '.' && p[1] == '/')
+				p += 2;
+
 			for(i = 0; extract_names[i] != 0; i++) {
-				if (strcmp(extract_names[i], file_entry->name) == 0) {
+				if (strcmp(extract_names[i], p) == 0) {
 					found_flag = TRUE;
 					break;
 				}
@@ -373,7 +383,7 @@ get_header_ar(FILE *src_stream)
 		/* raw_header[60] wont be '\n' as it should, but it doesnt matter */
 		memmove(ar.raw, &ar.raw[1], 59);
 	}
-		
+
 	typed = (file_header_t *) xcalloc(1, sizeof(file_header_t));
 
 	typed->size = (size_t) atoi(ar.formated.size);
@@ -409,7 +419,7 @@ get_header_ar(FILE *src_stream)
 	}
 	typed->name[strcspn(typed->name, " /")]='\0';
 
-	/* convert the rest of the now valid char header to its typed struct */	
+	/* convert the rest of the now valid char header to its typed struct */
 	parse_mode(ar.formated.mode, &typed->mode);
 	typed->mtime = atoi(ar.formated.date);
 	typed->uid = atoi(ar.formated.uid);
@@ -431,6 +441,42 @@ free_header_ar(file_header_t *ar_entry)
 	free(ar_entry);
 }
 
+static char uname_cache[32] = "";
+static uid_t uid_cache;
+
+static bool update_unamecache(char *uname) {
+	struct passwd *passwd;
+	if (!uname)
+		return FALSE;
+	if (!uname_cache[0] && strcmp(uname_cache, uname) == 0)
+		return TRUE;
+	passwd = getpwnam(uname);
+	if (passwd) {
+		uid_cache = passwd->pw_uid;
+		strncpy(uname, uname_cache, 32);
+		return TRUE;
+	}
+	return FALSE;
+}
+
+static char gname_cache[32] = "";
+static gid_t gid_cache;
+
+static bool update_gnamecache(char *gname) {
+	struct group *group;
+	if (!gname)
+		return FALSE;
+	if (!gname_cache[0] && strcmp(gname_cache, gname) == 0)
+		return TRUE;
+	group = getgrnam(gname);
+	if (group) {
+		gid_cache = group->gr_gid;
+		strncpy(gname, gname_cache, 32);
+		return TRUE;
+	}
+	return FALSE;
+}
+
 
 static file_header_t *
 get_header_tar(FILE *tar_stream)
@@ -466,7 +512,7 @@ get_header_tar(FILE *tar_stream)
 	}
 
 	if (fread(tar.raw, 1, 512, tar_stream) != 512) {
-		/* Unfortunatly its common for tar files to have all sorts of
+		/* Unfortunately its common for tar files to have all sorts of
 		 * trailing garbage, fail silently */
 //		error_msg("Couldnt read header");
 		return(NULL);
@@ -501,27 +547,7 @@ get_header_tar(FILE *tar_stream)
 	/* convert to type'ed variables */
 	tar_entry = xcalloc(1, sizeof(file_header_t));
 
-#ifdef CONFIG_FEATURE_TAR_GNU_EXTENSIONS
-        if (longname) {
-                tar_entry->name = longname;
-                longname = NULL;
-        }
-        else if (linkname) {
-                tar_entry->name = linkname;
-                linkname = NULL;
-        } else
-#endif
-        {
-                tar_entry->name = xstrndup(tar.formated.name, 100);
 
-                if (tar.formated.prefix[0]) {
-                        char *temp = tar_entry->name;
-                        char *prefixTemp = xstrndup(tar.formated.prefix, 155);
-                        tar_entry->name = concat_path_file(prefixTemp, temp);
-                        free(temp);
-                        free(prefixTemp);
-                }
-        }
 
 	// tar_entry->name = xstrdup(tar.formated.name);
 
@@ -530,11 +556,17 @@ get_header_tar(FILE *tar_stream)
 */
         tar_entry->mode = 07777 & strtol(tar.formated.mode, NULL, 8);
 
-	tar_entry->uid   = strtol(tar.formated.uid, NULL, 8);
-	tar_entry->gid   = strtol(tar.formated.gid, NULL, 8);
+	if (update_unamecache(tar.formated.uname))
+		tar_entry->uid = uid_cache;
+	else
+		tar_entry->uid = strtol(tar.formated.uid, NULL, 8);
+	if (update_gnamecache(tar.formated.gname))
+		tar_entry->gid = gid_cache;
+	else
+		tar_entry->gid = strtol(tar.formated.gid, NULL, 8);
 	tar_entry->size  = strtol(tar.formated.size, NULL, 8);
 	tar_entry->mtime = strtol(tar.formated.mtime, NULL, 8);
-	tar_entry->link_name  = *tar.formated.linkname != '\0' ? xstrndup(tar.formated.linkname, 100) : NULL;
+
 	tar_entry->device = (strtol(tar.formated.devmajor, NULL, 8) << 8) +
 		strtol(tar.formated.devminor, NULL, 8);
 
@@ -550,7 +582,7 @@ get_header_tar(FILE *tar_stream)
 # ifdef CONFIG_FEATURE_TAR_OLDGNU_COMPATABILITY
 		if (last_char_is(tar_entry->name, '/')) {
 			tar_entry->mode |= S_IFDIR;
-		} else 
+		} else
 # endif
 			tar_entry->mode |= S_IFREG;
 		break;
@@ -586,7 +618,6 @@ get_header_tar(FILE *tar_stream)
 			linkname[tar_entry->size] = '\0';
 			archive_offset += tar_entry->size;
 
-			tar_entry->name = linkname;
 			return(get_header_tar(tar_stream));
 		}
 	case 'D':
@@ -602,6 +633,34 @@ get_header_tar(FILE *tar_stream)
 
 	}
 
+
+#ifdef CONFIG_FEATURE_TAR_GNU_EXTENSIONS
+        if (longname) {
+                tar_entry->name = longname;
+                longname = NULL;
+        } else
+#endif
+        {
+                tar_entry->name = xstrndup(tar.formated.name, 100);
+
+                if (tar.formated.prefix[0]) {
+                        char *temp = tar_entry->name;
+                        char *prefixTemp = xstrndup(tar.formated.prefix, 155);
+                        tar_entry->name = concat_path_file(prefixTemp, temp);
+                        free(temp);
+                        free(prefixTemp);
+                }
+        }
+
+	if (linkname) {
+		tar_entry->link_name = linkname;
+		linkname = NULL;
+	} else
+	{
+		tar_entry->link_name = *tar.formated.linkname != '\0' ?
+			xstrndup(tar.formated.linkname, 100) : NULL;
+	}
+
 	return(tar_entry);
 }
 
@@ -612,14 +671,14 @@ free_header_tar(file_header_t *tar_entry)
 		return;
 
 	free(tar_entry->name);
-	if (tar_entry->link_name);
+	if (tar_entry->link_name)
 		free(tar_entry->link_name);
 
 	free(tar_entry);
 }
 
 char *
-deb_extract(const char *package_filename, FILE *out_stream, 
+deb_extract(const char *package_filename, FILE *out_stream,
 	const int extract_function, const char *prefix,
 	const char *filename, int *err)
 {
@@ -638,11 +697,11 @@ deb_extract(const char *package_filename, FILE *out_stream,
 		file_list[0] = filename;
 		file_list[1] = NULL;
 	}
-	
+
 	if (extract_function & extract_control_tar_gz) {
 		ared_file = "control.tar.gz";
 	}
-	else if (extract_function & extract_data_tar_gz) {		
+	else if (extract_function & extract_data_tar_gz) {
 		ared_file = "data.tar.gz";
 	} else {
                 opkg_msg(ERROR, "Internal error: extract_function=%x\n",
@@ -715,7 +774,7 @@ deb_extract(const char *package_filename, FILE *out_stream,
 			*err = -1;
 			goto cleanup;
 		}
-		
+
 		/* walk through outer tar file to find ared_file */
 		while ((tar_header = get_header_tar(unzipped_opkg_stream)) != NULL) {
                         int name_offset = 0;
@@ -732,12 +791,12 @@ deb_extract(const char *package_filename, FILE *out_stream,
 				}
 				archive_offset = 0;
 
-				output_buffer = unarchive(uncompressed_stream, 
-							  out_stream, 
+				output_buffer = unarchive(uncompressed_stream,
+							  out_stream,
 							  get_header_tar,
 							  free_header_tar,
-							  extract_function, 
-							  prefix, 
+							  extract_function,
+							  prefix,
 							  file_list,
 							  err);
 
