@@ -46,14 +46,16 @@ opkg_conf_t *conf = &_conf;
  */
 opkg_option_t options[] = {
 	  { "cache", OPKG_OPT_TYPE_STRING, &_conf.cache},
+	  { "lists_dir", OPKG_OPT_TYPE_STRING, &_conf.lists_dir },
 	  { "force_defaults", OPKG_OPT_TYPE_BOOL, &_conf.force_defaults },
-          { "force_maintainer", OPKG_OPT_TYPE_BOOL, &_conf.force_maintainer }, 
+          { "force_maintainer", OPKG_OPT_TYPE_BOOL, &_conf.force_maintainer },
 	  { "force_depends", OPKG_OPT_TYPE_BOOL, &_conf.force_depends },
 	  { "force_overwrite", OPKG_OPT_TYPE_BOOL, &_conf.force_overwrite },
 	  { "force_downgrade", OPKG_OPT_TYPE_BOOL, &_conf.force_downgrade },
 	  { "force_reinstall", OPKG_OPT_TYPE_BOOL, &_conf.force_reinstall },
 	  { "force_space", OPKG_OPT_TYPE_BOOL, &_conf.force_space },
-          { "check_signature", OPKG_OPT_TYPE_BOOL, &_conf.check_signature }, 
+	  { "force_postinstall", OPKG_OPT_TYPE_BOOL, &_conf.force_postinstall },
+          { "check_signature", OPKG_OPT_TYPE_BOOL, &_conf.check_signature },
 	  { "ftp_proxy", OPKG_OPT_TYPE_STRING, &_conf.ftp_proxy },
 	  { "http_proxy", OPKG_OPT_TYPE_STRING, &_conf.http_proxy },
 	  { "no_proxy", OPKG_OPT_TYPE_STRING, &_conf.no_proxy },
@@ -62,17 +64,19 @@ opkg_option_t options[] = {
 	  { "download_only", OPKG_OPT_TYPE_BOOL, &_conf.download_only },
 	  { "nodeps", OPKG_OPT_TYPE_BOOL, &_conf.nodeps },
 	  { "offline_root", OPKG_OPT_TYPE_STRING, &_conf.offline_root },
+	  { "overlay_root", OPKG_OPT_TYPE_STRING, &_conf.overlay_root },
 	  { "proxy_passwd", OPKG_OPT_TYPE_STRING, &_conf.proxy_passwd },
 	  { "proxy_user", OPKG_OPT_TYPE_STRING, &_conf.proxy_user },
 	  { "query-all", OPKG_OPT_TYPE_BOOL, &_conf.query_all },
 	  { "tmp_dir", OPKG_OPT_TYPE_STRING, &_conf.tmp_dir },
 	  { "verbosity", OPKG_OPT_TYPE_INT, &_conf.verbosity },
+          { "combine", OPKG_OPT_TYPE_BOOL, &_conf.combine },
 #if defined(HAVE_OPENSSL)
 	  { "signature_ca_file", OPKG_OPT_TYPE_STRING, &_conf.signature_ca_file },
 	  { "signature_ca_path", OPKG_OPT_TYPE_STRING, &_conf.signature_ca_path },
 #endif
 #if defined(HAVE_PATHFINDER)
-          { "check_x509_path", OPKG_OPT_TYPE_BOOL, &_conf.check_x509_path }, 
+          { "check_x509_path", OPKG_OPT_TYPE_BOOL, &_conf.check_x509_path },
 #endif
 #if defined(HAVE_SSLCURL) && defined(HAVE_CURL)
           { "ssl_engine", OPKG_OPT_TYPE_STRING, &_conf.ssl_engine },
@@ -89,15 +93,15 @@ opkg_option_t options[] = {
 };
 
 static int
-resolve_pkg_dest_list(nv_pair_list_t *nv_pair_list)
+resolve_pkg_dest_list(void)
 {
      nv_pair_list_elt_t *iter;
      nv_pair_t *nv_pair;
      pkg_dest_t *dest;
      char *root_dir;
 
-     for (iter = nv_pair_list_first(nv_pair_list); iter;
-		     iter = nv_pair_list_next(nv_pair_list, iter)) {
+     for (iter = nv_pair_list_first(&conf->tmp_dest_list); iter;
+		     iter = nv_pair_list_next(&conf->tmp_dest_list, iter)) {
 	  nv_pair = (nv_pair_t *)iter->data;
 
 	  if (conf->offline_root) {
@@ -156,7 +160,7 @@ opkg_conf_set_option(const char *name, const char *value)
 			 opkg_msg(ERROR, "Option %s needs an argument\n",
 				name);
 			 return -1;
-		    }		    
+		    }
 	       case OPKG_OPT_TYPE_STRING:
 		    if (value) {
 			    if (*(char **)options[i].value) {
@@ -176,7 +180,7 @@ opkg_conf_set_option(const char *name, const char *value)
 	  }
 	  i++;
      }
-    
+
      opkg_msg(ERROR, "Unrecognized option: %s=%s\n", name, value);
      return -1;
 }
@@ -184,13 +188,13 @@ opkg_conf_set_option(const char *name, const char *value)
 static int
 opkg_conf_parse_file(const char *filename,
 				pkg_src_list_t *pkg_src_list,
-				nv_pair_list_t *tmp_dest_nv_pair_list)
+				pkg_src_list_t *dist_src_list)
 {
      int line_num = 0;
      int err = 0;
      FILE *file;
      regex_t valid_line_re, comment_re;
-#define regmatch_size 12
+#define regmatch_size 14
      regmatch_t regmatch[regmatch_size];
 
      file = fopen(filename, "r");
@@ -202,7 +206,7 @@ opkg_conf_parse_file(const char *filename,
 
      opkg_msg(INFO, "Loading conf file %s.\n", filename);
 
-     err = xregcomp(&comment_re, 
+     err = xregcomp(&comment_re,
 		    "^[[:space:]]*(#.*|[[:space:]]*)$",
 		    REG_EXTENDED);
      if (err)
@@ -212,7 +216,7 @@ opkg_conf_parse_file(const char *filename,
 		     "^[[:space:]]*(\"([^\"]*)\"|([^[:space:]]*))"
 		     "[[:space:]]*(\"([^\"]*)\"|([^[:space:]]*))"
 		     "[[:space:]]*(\"([^\"]*)\"|([^[:space:]]*))"
-		     "([[:space:]]+([^[:space:]]+))?[[:space:]]*$",
+		     "([[:space:]]+([^[:space:]]+))?([[:space:]]+(.*))?[[:space:]]*$",
 		     REG_EXTENDED);
      if (err)
 	  goto err2;
@@ -263,11 +267,20 @@ opkg_conf_parse_file(const char *filename,
 
 	  extra = NULL;
 	  if (regmatch[11].rm_so > 0) {
+	     if (regmatch[13].rm_so > 0 && regmatch[13].rm_so!=regmatch[13].rm_eo )
+	       extra = xstrndup (line + regmatch[11].rm_so,
+				regmatch[13].rm_eo - regmatch[11].rm_so);
+	     else
 	       extra = xstrndup (line + regmatch[11].rm_so,
 				regmatch[11].rm_eo - regmatch[11].rm_so);
 	  }
 
-	  /* We use the tmp_dest_nv_pair_list below instead of
+	  if (regmatch[13].rm_so!=regmatch[13].rm_eo && strncmp(type, "dist", 4)!=0) {
+	       opkg_msg(ERROR, "%s:%d: Ignoring config line with trailing garbage: `%s'\n",
+		       filename, line_num, line);
+	  } else {
+
+	  /* We use the conf->tmp_dest_list below instead of
 	     conf->pkg_dest_list because we might encounter an
 	     offline_root option later and that would invalidate the
 	     directories we would have computed in
@@ -275,6 +288,20 @@ opkg_conf_parse_file(const char *filename,
 	     tmp_src_nv_pair_list for sake of symmetry.) */
 	  if (strcmp(type, "option") == 0) {
 	       opkg_conf_set_option(name, value);
+ 	  } else if (strcmp(type, "dist") == 0) {
+ 	       if (!nv_pair_list_find((nv_pair_list_t*) dist_src_list, name)) {
+ 		    pkg_src_list_append (dist_src_list, name, value, extra, 0);
+ 	       } else {
+ 		    opkg_msg(ERROR, "Duplicate dist declaration (%s %s). "
+ 				    "Skipping.\n", name, value);
+ 	       }
+ 	  } else if (strcmp(type, "dist/gz") == 0) {
+ 	       if (!nv_pair_list_find((nv_pair_list_t*) dist_src_list, name)) {
+ 		    pkg_src_list_append (dist_src_list, name, value, extra, 1);
+ 	       } else {
+ 		    opkg_msg(ERROR, "Duplicate dist declaration (%s %s). "
+ 				    "Skipping.\n", name, value);
+ 	       }
 	  } else if (strcmp(type, "src") == 0) {
 	       if (!nv_pair_list_find((nv_pair_list_t*) pkg_src_list, name)) {
 		    pkg_src_list_append (pkg_src_list, name, value, extra, 0);
@@ -290,8 +317,11 @@ opkg_conf_parse_file(const char *filename,
 				   "Skipping.\n", name, value);
 	       }
 	  } else if (strcmp(type, "dest") == 0) {
-	       nv_pair_list_append(tmp_dest_nv_pair_list, name, value);
+	       nv_pair_list_append(&conf->tmp_dest_list, name, value);
 	  } else if (strcmp(type, "lists_dir") == 0) {
+               /* Ensure we don't leak memory if lists_dir has already been set.
+                */
+               free(conf->lists_dir);
 	       conf->lists_dir = xstrdup(value);
 	  } else if (strcmp(type, "arch") == 0) {
 	       opkg_msg(INFO, "Supported arch %s priority (%s)\n", name, value);
@@ -306,6 +336,8 @@ opkg_conf_parse_file(const char *filename,
 		       filename, line_num, line);
 	  }
 
+	  }
+
 	  free(type);
 	  free(name);
 	  free(value);
@@ -359,7 +391,8 @@ opkg_conf_write_status_files(void)
 	  /* We don't need most uninstalled packages in the status file */
 	  if (pkg->state_status == SS_NOT_INSTALLED
 	      && (pkg->state_want == SW_UNKNOWN
-		  || pkg->state_want == SW_DEINSTALL
+		  || (pkg->state_want == SW_DEINSTALL
+			  && pkg->state_flag != SF_HOLD)
 		  || pkg->state_want == SW_PURGE)) {
 	       continue;
 	  }
@@ -408,10 +441,21 @@ glob_errfunc(const char *epath, int eerrno)
 
 int
 opkg_conf_init(void)
+{
+	pkg_src_list_init(&conf->pkg_src_list);
+	pkg_src_list_init(&conf->dist_src_list);
+	pkg_dest_list_init(&conf->pkg_dest_list);
+	pkg_dest_list_init(&conf->tmp_dest_list);
+	nv_pair_list_init(&conf->arch_list);
+
+	return 0;
+}
+
+int
+opkg_conf_load(void)
 {
 	int i, glob_ret;
 	char *tmp, *tmp_dir_base, **tmp_val;
-	nv_pair_list_t tmp_dest_nv_pair_list;
 	glob_t globbuf;
 	char *etc_opkg_conf_pattern;
 
@@ -421,11 +465,6 @@ opkg_conf_init(void)
 	conf->check_x509_path = 1;
 #endif
 
-	pkg_src_list_init(&conf->pkg_src_list);
-	pkg_dest_list_init(&conf->pkg_dest_list);
-	nv_pair_list_init(&conf->arch_list);
-	nv_pair_list_init(&tmp_dest_nv_pair_list);
-
 	if (!conf->offline_root)
 		conf->offline_root = xstrdup(getenv("OFFLINE_ROOT"));
 
@@ -436,42 +475,43 @@ opkg_conf_init(void)
 			goto err0;
 		}
 		if (opkg_conf_parse_file(conf->conf_file,
-				&conf->pkg_src_list, &tmp_dest_nv_pair_list))
+				&conf->pkg_src_list, &conf->dist_src_list))
 			goto err1;
 	}
-
-	if (conf->offline_root)
-		sprintf_alloc(&etc_opkg_conf_pattern, "%s/etc/opkg/*.conf", conf->offline_root);
 	else {
-		const char *conf_file_dir = getenv("OPKG_CONF_DIR");
-		if (conf_file_dir == NULL)
-			conf_file_dir = OPKG_CONF_DEFAULT_CONF_FILE_DIR;
-			sprintf_alloc(&etc_opkg_conf_pattern, "%s/*.conf", conf_file_dir);
-	}
+		if (conf->offline_root)
+			sprintf_alloc(&etc_opkg_conf_pattern, "%s/etc/opkg/*.conf", conf->offline_root);
+		else {
+			const char *conf_file_dir = getenv("OPKG_CONF_DIR");
+			if (conf_file_dir == NULL)
+				conf_file_dir = OPKG_CONF_DEFAULT_CONF_FILE_DIR;
+				sprintf_alloc(&etc_opkg_conf_pattern, "%s/*.conf", conf_file_dir);
+		}
 
-	memset(&globbuf, 0, sizeof(globbuf));
-	glob_ret = glob(etc_opkg_conf_pattern, 0, glob_errfunc, &globbuf);
-	if (glob_ret && glob_ret != GLOB_NOMATCH) {
-		free(etc_opkg_conf_pattern);
-		globfree(&globbuf);
-		goto err1;
-	}
-
-	free(etc_opkg_conf_pattern);
-
-	for (i = 0; i < globbuf.gl_pathc; i++) {
-		if (globbuf.gl_pathv[i]) 
-			if (conf->conf_file &&
-					!strcmp(conf->conf_file, globbuf.gl_pathv[i]))
-				continue;
-		if ( opkg_conf_parse_file(globbuf.gl_pathv[i], 
-			&conf->pkg_src_list, &tmp_dest_nv_pair_list)<0) {
+		memset(&globbuf, 0, sizeof(globbuf));
+		glob_ret = glob(etc_opkg_conf_pattern, 0, glob_errfunc, &globbuf);
+		if (glob_ret && glob_ret != GLOB_NOMATCH) {
+			free(etc_opkg_conf_pattern);
 			globfree(&globbuf);
 			goto err1;
 		}
-	}
 
-	globfree(&globbuf);
+		free(etc_opkg_conf_pattern);
+
+		for (i = 0; i < globbuf.gl_pathc; i++) {
+			if (globbuf.gl_pathv[i])
+				if (conf->conf_file &&
+						!strcmp(conf->conf_file, globbuf.gl_pathv[i]))
+					continue;
+			if ( opkg_conf_parse_file(globbuf.gl_pathv[i],
+				&conf->pkg_src_list, &conf->dist_src_list)<0) {
+				globfree(&globbuf);
+				goto err1;
+			}
+		}
+
+		globfree(&globbuf);
+	}
 
 	if (conf->offline_root)
 		sprintf_alloc (&lock_file, "%s/%s", conf->offline_root, OPKGLOCKFILE);
@@ -486,7 +526,11 @@ opkg_conf_init(void)
 
 	if (lockf(lock_fd, F_TLOCK, (off_t)0) == -1) {
 		opkg_perror(ERROR, "Could not lock %s", lock_file);
-		goto err3;
+		if (close(lock_fd) == -1)
+			opkg_perror(ERROR, "Couldn't close descriptor %d (%s)",
+				lock_fd, lock_file);
+		lock_fd = -1;
+		goto err2;
 	}
 
 	if (conf->tmp_dir)
@@ -502,7 +546,7 @@ opkg_conf_init(void)
 	conf->tmp_dir = mkdtemp(tmp);
 	if (conf->tmp_dir == NULL) {
 		opkg_perror(ERROR, "Creating temp dir %s failed", tmp);
-		goto err4;
+		goto err3;
 	}
 
 	pkg_hash_init();
@@ -526,42 +570,44 @@ opkg_conf_init(void)
 	}
 
 	/* Even if there is no conf file, we'll need at least one dest. */
-	if (nv_pair_list_empty(&tmp_dest_nv_pair_list)) {
-	nv_pair_list_append(&tmp_dest_nv_pair_list,
+	if (nv_pair_list_empty(&conf->tmp_dest_list)) {
+		nv_pair_list_append(&conf->tmp_dest_list,
 			OPKG_CONF_DEFAULT_DEST_NAME,
 			OPKG_CONF_DEFAULT_DEST_ROOT_DIR);
 	}
 
-	if (resolve_pkg_dest_list(&tmp_dest_nv_pair_list))
-		goto err5;
+	if (resolve_pkg_dest_list())
+		goto err4;
 
-	nv_pair_list_deinit(&tmp_dest_nv_pair_list);
+	nv_pair_list_deinit(&conf->tmp_dest_list);
 
 	return 0;
 
 
-err5:
-	free(conf->lists_dir);
-
+err4:
 	pkg_hash_deinit();
 	hash_table_deinit(&conf->file_hash);
 	hash_table_deinit(&conf->obs_file_hash);
 
 	if (rmdir(conf->tmp_dir) == -1)
 		opkg_perror(ERROR, "Couldn't remove dir %s", conf->tmp_dir);
-err4:
+err3:
 	if (lockf(lock_fd, F_ULOCK, (off_t)0) == -1)
 		opkg_perror(ERROR, "Couldn't unlock %s", lock_file);
-err3:
+
 	if (close(lock_fd) == -1)
 		opkg_perror(ERROR, "Couldn't close descriptor %d (%s)",
 				lock_fd, lock_file);
 	if (unlink(lock_file) == -1)
 		opkg_perror(ERROR, "Couldn't unlink %s", lock_file);
 err2:
-	free(lock_file);
+	if (lock_file) {
+		free(lock_file);
+		lock_file = NULL;
+	}
 err1:
 	pkg_src_list_deinit(&conf->pkg_src_list);
+	pkg_src_list_deinit(&conf->dist_src_list);
 	pkg_dest_list_deinit(&conf->pkg_dest_list);
 	nv_pair_list_deinit(&conf->arch_list);
 
@@ -575,7 +621,7 @@ err1:
 		}
 	}
 err0:
-	nv_pair_list_deinit(&tmp_dest_nv_pair_list);
+	nv_pair_list_deinit(&conf->tmp_dest_list);
 	if (conf->dest_str)
 		free(conf->dest_str);
 	if (conf->conf_file)
@@ -590,9 +636,8 @@ opkg_conf_deinit(void)
 	int i;
 	char **tmp;
 
-	rm_r(conf->tmp_dir);
-
-	free(conf->lists_dir);
+	if (conf->tmp_dir)
+		rm_r(conf->tmp_dir);
 
 	if (conf->dest_str)
 		free(conf->dest_str);
@@ -601,6 +646,7 @@ opkg_conf_deinit(void)
 		free(conf->conf_file);
 
 	pkg_src_list_deinit(&conf->pkg_src_list);
+	pkg_src_list_deinit(&conf->dist_src_list);
 	pkg_dest_list_deinit(&conf->pkg_dest_list);
 	nv_pair_list_deinit(&conf->arch_list);
 
@@ -614,7 +660,7 @@ opkg_conf_deinit(void)
 		}
 	}
 
-	if (conf->verbosity >= DEBUG) { 
+	if (conf->verbosity >= DEBUG) {
 		hash_print_stats(&conf->pkg_hash);
 		hash_print_stats(&conf->file_hash);
 		hash_print_stats(&conf->obs_file_hash);
@@ -624,15 +670,20 @@ opkg_conf_deinit(void)
 	hash_table_deinit(&conf->file_hash);
 	hash_table_deinit(&conf->obs_file_hash);
 
-	if (lockf(lock_fd, F_ULOCK, (off_t)0) == -1)
-		opkg_perror(ERROR, "Couldn't unlock %s", lock_file);
+	if (lock_fd != -1) {
+		if (lockf(lock_fd, F_ULOCK, (off_t)0) == -1)
+			opkg_perror(ERROR, "Couldn't unlock %s", lock_file);
 
-	if (close(lock_fd) == -1)
-		opkg_perror(ERROR, "Couldn't close descriptor %d (%s)",
-				lock_fd, lock_file);
+		if (close(lock_fd) == -1)
+			opkg_perror(ERROR, "Couldn't close descriptor %d (%s)",
+					lock_fd, lock_file);
 
-	if (unlink(lock_file) == -1)
-		opkg_perror(ERROR, "Couldn't unlink %s", lock_file);
+	}
 
-	free(lock_file);
+	if (lock_file) {
+		if (unlink(lock_file) == -1)
+			opkg_perror(ERROR, "Couldn't unlink %s", lock_file);
+
+		free(lock_file);
+	}
 }
