@@ -19,10 +19,18 @@
 #define OPKG_INSTALL_H
 
 #include "pkg.h"
-#include "opkg_conf.h"
-#include "opkg_error.h"
-
-int opkg_install_by_name(opkg_conf_t *conf, const char *pkg_name);
-int opkg_install_pkg(opkg_conf_t *conf, pkg_t *pkg, int from_upgrading);
+#include "str_list.h"
 
+#ifdef __cplusplus
+extern "C" {
 #endif
+
+int opkg_install_by_name(const char *pkg_name);
+int opkg_install_multiple_by_name(str_list_t *pkg_names);
+int opkg_install_pkg(pkg_t *pkg, int from_upgrading);
+
+#ifdef __cplusplus
+}
+#endif
+
+#endif /* OPKG_INSTALL_H */
