@@ -16,15 +16,23 @@
    General Public License for more details.
 */
 
+#include "config.h"
+
 #include <stdio.h>
 #include <stdlib.h>
 
 #include "opkg_install.h"
 #include "opkg_upgrade.h"
 #include "opkg_message.h"
+#include "libbb/libbb.h"
 
-int
-opkg_upgrade_pkg(pkg_t *old)
+typedef struct upgrade_pair {
+    pkg_t * old;
+    pkg_t * new;
+} upgrade_pair_t;
+
+static int
+opkg_prepare_upgrade_pkg(pkg_t *old, pkg_t **pkg)
 {
      pkg_t *new;
      int cmp;
@@ -44,10 +52,10 @@ opkg_upgrade_pkg(pkg_t *old)
           free(old_version);
           return 0;
      }
-          
+
      old_version = pkg_version_str_alloc(old);
      new_version = pkg_version_str_alloc(new);
-               
+
      cmp = pkg_compare_versions(old, new);
      opkg_msg(DEBUG, "Comparing visible versions of pkg %s:"
                   "\n\t%s is installed "
@@ -74,9 +82,92 @@ opkg_upgrade_pkg(pkg_t *old)
     free(old_version);
     free(new_version);
     new->state_flag |= SF_USER;
-    return opkg_install_pkg(new,1);
+    new->state_want = SW_INSTALL;
+
+    *pkg = new;
+    return 1;
 }
 
+int
+opkg_upgrade_pkg(pkg_t *old)
+{
+    pkg_t *new;
+    int r;
+
+    r = opkg_prepare_upgrade_pkg(old, &new);
+    if (r <= 0)
+        return r;
+
+    opkg_msg(DEBUG2,"Calling opkg_install_pkg for %s %s.\n",
+             new->name, new->version);
+    r = opkg_install_pkg(new,1);
+    if (r < 0) {
+        /* The installation failed so we need to reset the appropriate
+         * state_want flags.
+         */
+        new->state_want = SW_UNKNOWN;
+        old->state_want = SW_INSTALL;
+    }
+
+    return r;
+}
+
+int
+opkg_upgrade_multiple_pkgs(pkg_vec_t *pkgs_to_upgrade)
+{
+    int r;
+    unsigned int i;
+    pkg_t *pkg;
+    pkg_t *new;
+    upgrade_pair_t *pair;
+    void_list_t upgrade_pairs;
+    void_list_elt_t *iter;
+    int errors = 0;
+
+    void_list_init(&upgrade_pairs);
+
+    /* Prepare all packages. */
+    for (i = 0; i < pkgs_to_upgrade->len; i++) {
+        pkg = pkgs_to_upgrade->pkgs[i];
+
+        r = opkg_prepare_upgrade_pkg(pkg, &new);
+        if (r < 0)
+            errors++;
+        if (r <= 0)
+            continue;
+
+        pair = (upgrade_pair_t *)xmalloc(sizeof(upgrade_pair_t));
+        pair->old = pkg;
+        pair->new = new;
+        void_list_push(&upgrade_pairs, pair);
+    }
+
+    /* Install all new packages. */
+    list_for_each_entry(iter, &upgrade_pairs.head, node) {
+        pair = iter->data;
+        pkg = pair->new;
+
+        opkg_msg(DEBUG2,"Calling opkg_install_pkg for %s %s.\n",
+                 pkg->name, pkg->version);
+        r = opkg_install_pkg(pkg, 1);
+        if (r < 0) {
+            /* The installation failed so we need to reset the appropriate
+             * state_want flags.
+             */
+            pkg->state_want = SW_UNKNOWN;
+            pair->old->state_want = SW_INSTALL;
+            errors++;
+        }
+
+        free(pair);
+    }
+
+    void_list_deinit(&upgrade_pairs);
+    if (errors > 0)
+        return -1;
+
+    return 0;
+}
 
 static void
 pkg_hash_check_installed_pkg_helper(const char *pkg_name, void *entry,
@@ -122,7 +213,7 @@ prepare_upgrade_list(void)
         cmp = pkg_compare_versions(old, new);
 
         if ( cmp < 0 ) {
-           node = active_list_move_node(all, head, &old->list); 
+           node = active_list_move_node(all, head, &old->list);
         }
     }
     active_list_head_delete(all);
