@@ -21,6 +21,7 @@
 #include <string.h>
 #include <ctype.h>
 #include <unistd.h>
+#include <libgen.h>
 
 #include "pkg.h"
 
@@ -78,6 +79,7 @@ pkg_init(pkg_t *pkg)
      pkg->epoch = 0;
      pkg->version = NULL;
      pkg->revision = NULL;
+     pkg->force_reinstall = 0;
      pkg->dest = NULL;
      pkg->src = NULL;
      pkg->architecture = NULL;
@@ -85,6 +87,7 @@ pkg_init(pkg_t *pkg)
      pkg->section = NULL;
      pkg->description = NULL;
      pkg->state_want = SW_UNKNOWN;
+     pkg->wanted_by = pkg_vec_alloc();
      pkg->state_flag = SF_OK;
      pkg->state_status = SS_NOT_INSTALLED;
      pkg->depends_str = NULL;
@@ -95,7 +98,7 @@ pkg_init(pkg_t *pkg)
      pkg->recommends_str = NULL;
      pkg->suggests_count = 0;
      pkg->recommends_count = 0;
-     
+
      active_list_init(&pkg->list);
 
      pkg->conflicts = NULL;
@@ -103,7 +106,7 @@ pkg_init(pkg_t *pkg)
 
      pkg->replaces = NULL;
      pkg->replaces_count = 0;
-    
+
      pkg->pre_depends_count = 0;
      pkg->pre_depends_str = NULL;
      pkg->provides_count = 0;
@@ -168,6 +171,8 @@ pkg_deinit(pkg_t *pkg)
 	/* revision shares storage with version, so don't free */
 	pkg->revision = NULL;
 
+	pkg->force_reinstall = 0;
+
 	/* owned by opkg_conf_t */
 	pkg->dest = NULL;
 	/* owned by opkg_conf_t */
@@ -188,8 +193,9 @@ pkg_deinit(pkg_t *pkg)
 	if (pkg->description)
 		free(pkg->description);
 	pkg->description = NULL;
-	
+
 	pkg->state_want = SW_UNKNOWN;
+	pkg_vec_free(pkg->wanted_by);
 	pkg->state_flag = SF_OK;
 	pkg->state_status = SS_NOT_INSTALLED;
 
@@ -221,11 +227,11 @@ pkg_deinit(pkg_t *pkg)
 
 	pkg->pre_depends_count = 0;
 	pkg->provides_count = 0;
-	
+
 	if (pkg->filename)
 		free(pkg->filename);
 	pkg->filename = NULL;
-	
+
 	if (pkg->local_filename)
 		free(pkg->local_filename);
 	pkg->local_filename = NULL;
@@ -281,7 +287,7 @@ pkg_init_from_file(pkg_t *pkg, const char *filename)
 	pkg->local_filename = xstrdup(filename);
 
 	tmp = xstrdup(filename);
-	sprintf_alloc(&control_path, "%s/%s.control.XXXXXX", 
+	sprintf_alloc(&control_path, "%s/%s.control.XXXXXX",
                         conf->tmp_dir,
                         basename(tmp));
 	free(tmp);
@@ -374,10 +380,8 @@ pkg_merge(pkg_t *oldpkg, pkg_t *newpkg)
 	  oldpkg->provides_count = newpkg->provides_count;
 	  newpkg->provides_count = 0;
 
-	  if (!oldpkg->provides) {
-		oldpkg->provides = newpkg->provides;
-		newpkg->provides = NULL;
-	  }
+	  oldpkg->provides = newpkg->provides;
+	  newpkg->provides = NULL;
      }
 
      if (!oldpkg->conflicts_count) {
@@ -461,7 +465,7 @@ set_flags_from_control(pkg_t *pkg){
 
      fp = fopen(file_name, "r");
      if (fp == NULL) {
-	     opkg_perror(ERROR, "Failed to open %s");
+	     opkg_perror(ERROR, "Failed to open %s", file_name);
 	     free(file_name);
 	     return;
      }
@@ -641,8 +645,8 @@ pkg_formatted_field(FILE *fp, pkg_t *pkg, const char *field)
                fprintf(fp, "Conffiles:\n");
 	       for (iter = nv_pair_list_first(&pkg->conffiles); iter; iter = nv_pair_list_next(&pkg->conffiles, iter)) {
 		    if (((conffile_t *)iter->data)->name && ((conffile_t *)iter->data)->value) {
-                         fprintf(fp, " %s %s\n", 
-                                 ((conffile_t *)iter->data)->name, 
+                         fprintf(fp, " %s %s\n",
+                                 ((conffile_t *)iter->data)->name,
                                  ((conffile_t *)iter->data)->value);
 		    }
 	       }
@@ -713,7 +717,7 @@ pkg_formatted_field(FILE *fp, pkg_t *pkg, const char *field)
      case 'M':
 	  if (strcasecmp(field, "Maintainer") == 0) {
 	       if (pkg->maintainer) {
-                   fprintf(fp, "maintainer: %s\n", pkg->maintainer);
+                   fprintf(fp, "Maintainer: %s\n", pkg->maintainer);
 	       }
 	  } else if (strcasecmp(field, "MD5sum") == 0) {
 	       if (pkg->md5sum) {
@@ -730,7 +734,8 @@ pkg_formatted_field(FILE *fp, pkg_t *pkg, const char *field)
 	  } else if (strcasecmp(field, "Priority") == 0) {
                fprintf(fp, "Priority: %s\n", pkg->priority);
 	  } else if (strcasecmp(field, "Provides") == 0) {
-	       if (pkg->provides_count) {
+	       /* Don't print provides if this package provides only itself */
+	       if (pkg->provides_count > 1) {
                   fprintf(fp, "Provides:");
 		  for(i = 1; i < pkg->provides_count; i++) {
                       fprintf(fp, "%s %s", i == 1 ? "" : ",",
@@ -934,30 +939,35 @@ verrevcmp(const char *val, const char *ref) {
   return 0;
 }
 
+int
+pkg_compare_versions_no_reinstall(const pkg_t *pkg, const pkg_t *ref_pkg)
+{
+    int r;
+
+    r = pkg->epoch - ref_pkg->epoch;
+    if (r)
+        return r;
+
+    r = verrevcmp(pkg->version, ref_pkg->version);
+    if (r)
+        return r;
+
+    r = verrevcmp(pkg->revision, ref_pkg->revision);
+    return r;
+}
+
 int
 pkg_compare_versions(const pkg_t *pkg, const pkg_t *ref_pkg)
 {
-     int r;
+    int r;
 
-     if (pkg->epoch > ref_pkg->epoch) {
-	  return 1;
-     }
+    r = pkg_compare_versions_no_reinstall(pkg, ref_pkg);
+    if (r)
+        return r;
 
-     if (pkg->epoch < ref_pkg->epoch) {
-	  return -1;
-     }
-
-     r = verrevcmp(pkg->version, ref_pkg->version);
-     if (r) {
-	  return r;
-     }
-
-     r = verrevcmp(pkg->revision, ref_pkg->revision);
-     if (r) {
-	  return r;
-     }
-
-     return r;
+    /* Compare force_reinstall flags. */
+    r = pkg->force_reinstall - ref_pkg->force_reinstall;
+    return r;
 }
 
 
@@ -967,28 +977,24 @@ pkg_version_satisfied(pkg_t *it, pkg_t *ref, const char *op)
      int r;
 
      r = pkg_compare_versions(it, ref);
+     char *op2 = op;
+     enum version_constraint constraint = str_to_constraint(&op2);
 
-     if (strcmp(op, "<=") == 0 || strcmp(op, "<") == 0) {
-	  return r <= 0;
+     switch (constraint)
+     {
+     case EARLIER_EQUAL:
+          return r <= 0;
+     case LATER_EQUAL:
+          return r >= 0;
+     case EARLIER:
+          return r < 0;
+     case LATER:
+          return r > 0;
+     case EQUAL:
+          return r == 0;
+     case NONE:
+          opkg_msg(ERROR, "Unknown operator: %s.\n", op);
      }
-
-     if (strcmp(op, ">=") == 0 || strcmp(op, ">") == 0) {
-	  return r >= 0;
-     }
-
-     if (strcmp(op, "<<") == 0) {
-	  return r < 0;
-     }
-
-     if (strcmp(op, ">>") == 0) {
-	  return r > 0;
-     }
-
-     if (strcmp(op, "=") == 0) {
-	  return r == 0;
-     }
-
-     opkg_msg(ERROR, "Unknown operator: %s.\n", op);
      return 0;
 }
 
@@ -1004,7 +1010,7 @@ pkg_name_version_and_architecture_compare(const void *p1, const void *p2)
 	       a->name, b->name);
        return 0;
      }
-       
+
      namecmp = strcmp(a->name, b->name);
      if (namecmp)
 	  return namecmp;
@@ -1072,6 +1078,7 @@ pkg_get_installed_files(pkg_t *pkg)
      char *line;
      char *installed_file_name;
      unsigned int rootdirlen = 0;
+     int list_from_package;
 
      pkg->installed_files_ref_cnt++;
 
@@ -1081,17 +1088,23 @@ pkg_get_installed_files(pkg_t *pkg)
 
      pkg->installed_files = str_list_alloc();
 
-     /* For uninstalled packages, get the file list directly from the package.
-	For installed packages, look at the package.list file in the database.
-     */
-     if (pkg->state_status == SS_NOT_INSTALLED || pkg->dest == NULL) {
+     /*
+      * For installed packages, look at the package.list file in the database.
+      * For uninstalled packages, get the file list directly from the package.
+      */
+     if (pkg->state_status == SS_NOT_INSTALLED || pkg->dest == NULL)
+	     list_from_package = 1;
+     else
+	     list_from_package = 0;
+
+     if (list_from_package) {
 	  if (pkg->local_filename == NULL) {
 	       return pkg->installed_files;
 	  }
 	  /* XXX: CLEANUP: Maybe rewrite this to avoid using a temporary
 	     file. In other words, change deb_extract so that it can
 	     simply return the file list as a char *[] rather than
-	     insisting on writing in to a FILE * as it does now. */
+	     insisting on writing it to a FILE * as it does now. */
 	  sprintf_alloc(&list_file_name, "%s/%s.list.XXXXXX",
 					  conf->tmp_dir, pkg->name);
 	  fd = mkstemp(list_file_name);
@@ -1140,14 +1153,14 @@ pkg_get_installed_files(pkg_t *pkg)
 
      while (1) {
 	  char *file_name;
-	
+
 	  line = file_read_line_alloc(list_file);
 	  if (line == NULL) {
 	       break;
 	  }
 	  file_name = line;
 
-	  if (pkg->state_status == SS_NOT_INSTALLED || pkg->dest == NULL) {
+	  if (list_from_package) {
 	       if (*file_name == '.') {
 		    file_name++;
 	       }
@@ -1173,7 +1186,7 @@ pkg_get_installed_files(pkg_t *pkg)
 
      fclose(list_file);
 
-     if (pkg->state_status == SS_NOT_INSTALLED || pkg->dest == NULL) {
+     if (list_from_package) {
 	  unlink(list_file_name);
 	  free(list_file_name);
      }
@@ -1246,9 +1259,7 @@ pkg_run_script(pkg_t *pkg, const char *script, const char *args)
      if (conf->noaction)
 	     return 0;
 
-     /* XXX: FEATURE: When conf->offline_root is set, we should run the
-	maintainer script within a chroot environment. */
-     if (conf->offline_root) {
+     if (conf->offline_root && !conf->force_postinstall) {
           opkg_msg(INFO, "Offline root mode: not running %s.%s.\n",
 			  pkg->name, script);
 	  return 0;
@@ -1291,7 +1302,9 @@ pkg_run_script(pkg_t *pkg, const char *script, const char *args)
      free(cmd);
 
      if (err) {
-	  opkg_msg(ERROR, "%s script returned status %d.\n", script, err);
+          if (!conf->offline_root)
+	       opkg_msg(ERROR, "package \"%s\" %s script returned status %d.\n",
+                    pkg->name, script, err);
 	  return err;
      }
 
@@ -1338,8 +1351,8 @@ pkg_info_preinstall_check(void)
 			       "files for pkg %s.\n", pkg->name);
 	       break;
 	  }
-	  for (iter = str_list_first(installed_files), niter = str_list_next(installed_files, iter); 
-                  iter; 
+	  for (iter = str_list_first(installed_files), niter = str_list_next(installed_files, iter);
+                  iter;
                   iter = niter, niter = str_list_next(installed_files, iter)) {
 	       char *installed_file = (char *) iter->data;
 	       file_hash_set_file_owner(installed_file, pkg);
