@@ -1,7 +1,7 @@
 /* pkg_depends.c - the opkg package management system
 
    Steven M. Ayer
-   
+
    Copyright (C) 2002 Compaq Computer Corporation
 
    This program is free software; you can redistribute it and/or
@@ -15,6 +15,8 @@
    General Public License for more details.
 */
 
+#include "config.h"
+
 #include <stdio.h>
 #include <ctype.h>
 
@@ -30,7 +32,6 @@ static int parseDepends(compound_depend_t *compound_depend, char * depend_str);
 static depend_t * depend_init(void);
 static char ** add_unresolved_dep(pkg_t * pkg, char ** the_lost, int ref_ndx);
 static char ** merge_unresolved(char ** oldstuff, char ** newstuff);
-static int is_pkg_in_pkg_vec(pkg_vec_t * vec, pkg_t * pkg);
 
 static int pkg_installed_and_constraint_satisfied(pkg_t *pkg, void *cdata)
 {
@@ -50,7 +51,7 @@ static int pkg_constraint_satisfied(pkg_t *pkg, void *cdata)
 	  return 0;
 }
 
-/* returns ndependencies or negative error value */ 
+/* returns ndependencies or negative error value */
 int
 pkg_hash_fetch_unsatisfied_dependencies(pkg_t * pkg, pkg_vec_t *unsatisfied,
 		char *** unresolved)
@@ -61,8 +62,8 @@ pkg_hash_fetch_unsatisfied_dependencies(pkg_t * pkg, pkg_vec_t *unsatisfied,
      char ** the_lost;
      abstract_pkg_t * ab_pkg;
 
-     /* 
-      * this is a setup to check for redundant/cyclic dependency checks, 
+     /*
+      * this is a setup to check for redundant/cyclic dependency checks,
       * which are marked at the abstract_pkg level
       */
      if (!(ab_pkg = pkg->parent)) {
@@ -71,9 +72,11 @@ pkg_hash_fetch_unsatisfied_dependencies(pkg_t * pkg, pkg_vec_t *unsatisfied,
 	  return 0;
      }
      if (ab_pkg->dependencies_checked) {    /* avoid duplicate or cyclic checks */
+	  opkg_msg(DEBUG2, "Already checked dependencies for '%s'.\n", ab_pkg->name);
 	  *unresolved = NULL;
 	  return 0;
-     } else { 
+     } else {
+	  opkg_msg(DEBUG2, "Checking dependencies for '%s'.\n", ab_pkg->name);
 	  ab_pkg->dependencies_checked = 1;  /* mark it for subsequent visits */
      }
      /**/
@@ -85,7 +88,7 @@ pkg_hash_fetch_unsatisfied_dependencies(pkg_t * pkg, pkg_vec_t *unsatisfied,
      }
 
      the_lost = NULL;
-	
+
      /* foreach dependency */
      for (i = 0; i < count; i++) {
 	  compound_depend_t * compound_depend = &pkg->depends[i];
@@ -100,7 +103,7 @@ pkg_hash_fetch_unsatisfied_dependencies(pkg_t * pkg, pkg_vec_t *unsatisfied,
 		    abstract_pkg_t *abpkg = possible_satisfiers[j]->pkg;
 		    abstract_pkg_vec_t *ab_provider_vec = abpkg->provided_by;
 		    int nposs = ab_provider_vec->len;
-		    abstract_pkg_t **ab_providers = ab_provider_vec->pkgs; 
+		    abstract_pkg_t **ab_providers = ab_provider_vec->pkgs;
 		    int l;
 		    for (l = 0; l < nposs; l++) {
 			 pkg_vec_t *test_vec = ab_providers[l]->pkgs;
@@ -108,7 +111,7 @@ pkg_hash_fetch_unsatisfied_dependencies(pkg_t * pkg, pkg_vec_t *unsatisfied,
 			 if (!test_vec){   /* no pkg_vec hooked up to the abstract_pkg!  (need another feed?) */
 			      continue;
 			 }
-	      
+
 			 /* cruise this possiblity's pkg_vec looking for an installed version */
 			 for (k = 0; k < test_vec->len; k++) {
 			      pkg_t *pkg_scout = test_vec->pkgs[k];
@@ -167,16 +170,10 @@ pkg_hash_fetch_unsatisfied_dependencies(pkg_t * pkg, pkg_vec_t *unsatisfied,
 	       /* foreach provided_by, which includes the abstract_pkg itself */
 	       depend_t *dependence_to_satisfy = possible_satisfiers[j];
 	       abstract_pkg_t *satisfying_apkg = possible_satisfiers[j]->pkg;
-	       pkg_t *satisfying_pkg = 
-		    pkg_hash_fetch_best_installation_candidate(satisfying_apkg, 
-							       pkg_installed_and_constraint_satisfied, 
+	       pkg_t *satisfying_pkg =
+		    pkg_hash_fetch_best_installation_candidate(satisfying_apkg,
+							       pkg_installed_and_constraint_satisfied,
 							       dependence_to_satisfy, 1);
-               /* Being that I can't test constraing in pkg_hash, I will test it here */
-	       if (satisfying_pkg != NULL) {
-                  if (!pkg_installed_and_constraint_satisfied ( satisfying_pkg,dependence_to_satisfy)) {
-	              satisfying_pkg = NULL;
-                  }
-               }
 	       opkg_msg(DEBUG, "satisfying_pkg=%p\n", satisfying_pkg);
 	       if (satisfying_pkg != NULL) {
 		    found = 1;
@@ -191,16 +188,10 @@ pkg_hash_fetch_unsatisfied_dependencies(pkg_t * pkg, pkg_vec_t *unsatisfied,
 		    /* foreach provided_by, which includes the abstract_pkg itself */
 		    depend_t *dependence_to_satisfy = possible_satisfiers[j];
 		    abstract_pkg_t *satisfying_apkg = possible_satisfiers[j]->pkg;
-		    pkg_t *satisfying_pkg = 
-			 pkg_hash_fetch_best_installation_candidate(satisfying_apkg, 
-								    pkg_constraint_satisfied, 
+		    pkg_t *satisfying_pkg =
+			 pkg_hash_fetch_best_installation_candidate(satisfying_apkg,
+								    pkg_constraint_satisfied,
 								    dependence_to_satisfy, 1);
-                    /* Being that I can't test constraing in pkg_hash, I will test it here too */
-	            if (satisfying_pkg != NULL) {
-                         if (!pkg_constraint_satisfied ( satisfying_pkg,dependence_to_satisfy)) {
-                            satisfying_pkg = NULL;
-                         }
-                    }
 
 		    /* user request overrides package recommendation */
 		    if (satisfying_pkg != NULL
@@ -239,7 +230,7 @@ pkg_hash_fetch_unsatisfied_dependencies(pkg_t * pkg, pkg_vec_t *unsatisfied,
 				pkg->name, satisfier_entry_pkg->name);
 		    } else {
 			 char ** newstuff = NULL;
-			 
+
 			 if (satisfier_entry_pkg != pkg &&
 			     !is_pkg_in_pkg_vec(unsatisfied, satisfier_entry_pkg)) {
 			      pkg_vec_insert(unsatisfied, satisfier_entry_pkg);
@@ -259,10 +250,91 @@ pkg_hash_fetch_unsatisfied_dependencies(pkg_t * pkg, pkg_vec_t *unsatisfied,
      return unsatisfied->len;
 }
 
-/*checking for conflicts !in replaces 
-  If a packages conflicts with another but is also replacing it, I should not consider it a 
-  really conflicts 
-  returns 0 if conflicts <> replaces or 1 if conflicts == replaces 
+
+pkg_vec_t *
+pkg_hash_fetch_satisfied_dependencies(pkg_t * pkg)
+{
+     pkg_vec_t *satisfiers;
+     int i, j, k;
+     int count;
+     abstract_pkg_t * ab_pkg;
+
+     satisfiers = pkg_vec_alloc();
+
+     /*
+      * this is a setup to check for redundant/cyclic dependency checks,
+      * which are marked at the abstract_pkg level
+      */
+     if (!(ab_pkg = pkg->parent)) {
+	  opkg_msg(ERROR, "Internal error, with pkg %s.\n", pkg->name);
+	  return satisfiers;
+     }
+
+     count = pkg->pre_depends_count + pkg->depends_count + pkg->recommends_count + pkg->suggests_count;
+     if (!count)
+	  return satisfiers;
+
+     /* foreach dependency */
+     for (i = 0; i < count; i++) {
+	  compound_depend_t * compound_depend = &pkg->depends[i];
+	  depend_t ** possible_satisfiers = compound_depend->possibilities;;
+
+          if (compound_depend->type == RECOMMEND || compound_depend->type == SUGGEST)
+              continue;
+
+	  if (compound_depend->type == GREEDY_DEPEND) {
+	       /* foreach possible satisfier */
+	       for (j = 0; j < compound_depend->possibility_count; j++) {
+		    /* foreach provided_by, which includes the abstract_pkg itself */
+		    abstract_pkg_t *abpkg = possible_satisfiers[j]->pkg;
+		    abstract_pkg_vec_t *ab_provider_vec = abpkg->provided_by;
+		    int nposs = ab_provider_vec->len;
+		    abstract_pkg_t **ab_providers = ab_provider_vec->pkgs;
+		    int l;
+		    for (l = 0; l < nposs; l++) {
+			 pkg_vec_t *test_vec = ab_providers[l]->pkgs;
+			 /* if no depends on this one, try the first package that Provides this one */
+			 if (!test_vec){   /* no pkg_vec hooked up to the abstract_pkg!  (need another feed?) */
+			      continue;
+			 }
+
+			 /* cruise this possiblity's pkg_vec looking for an installed version */
+			 for (k = 0; k < test_vec->len; k++) {
+			      pkg_t *pkg_scout = test_vec->pkgs[k];
+			      /* not installed, and not already known about? */
+			      if (pkg_scout->state_want == SW_INSTALL && pkg_scout != pkg)
+			          pkg_vec_insert(satisfiers, pkg_scout);
+			 }
+		    }
+	       }
+
+	       continue;
+	  }
+
+	  /* foreach possible satisfier, look for installed package  */
+	  for (j = 0; j < compound_depend->possibility_count; j++) {
+	       /* foreach provided_by, which includes the abstract_pkg itself */
+	       depend_t *dependence_to_satisfy = possible_satisfiers[j];
+	       abstract_pkg_t *satisfying_apkg = possible_satisfiers[j]->pkg;
+	       pkg_t *satisfying_pkg =
+		    pkg_hash_fetch_best_installation_candidate(satisfying_apkg,
+							       pkg_constraint_satisfied,
+							       dependence_to_satisfy, 0);
+	       if (satisfying_pkg != NULL
+                       && satisfying_pkg != pkg
+                       && (satisfying_pkg->state_want == SW_INSTALL
+                                || satisfying_pkg->state_want == SW_UNKNOWN))
+	           pkg_vec_insert(satisfiers, satisfying_pkg);
+	  }
+     }
+     return satisfiers;
+}
+
+
+/*checking for conflicts !in replaces
+  If a packages conflicts with another but is also replacing it, I should not consider it a
+  really conflicts
+  returns 0 if conflicts <> replaces or 1 if conflicts == replaces
 */
 static int
 is_pkg_a_replaces(pkg_t *pkg_scout,pkg_t *pkg)
@@ -297,11 +369,11 @@ pkg_vec_t * pkg_hash_fetch_conflicts(pkg_t * pkg)
     int i, j, k;
     int count;
     abstract_pkg_t * ab_pkg;
-    pkg_t **pkg_scouts; 
-    pkg_t *pkg_scout; 
+    pkg_t **pkg_scouts;
+    pkg_t *pkg_scout;
 
-    /* 
-     * this is a setup to check for redundant/cyclic dependency checks, 
+    /*
+     * this is a setup to check for redundant/cyclic dependency checks,
      * which are marked at the abstract_pkg level
      */
     if(!(ab_pkg = pkg->parent)){
@@ -340,7 +412,7 @@ pkg_vec_t * pkg_hash_fetch_conflicts(pkg_t * pkg)
                     pkg_scout = pkg_scouts[k];
                     if (!pkg_scout) {
                         opkg_msg(ERROR,  "Internal error: pkg_scout=NULL\n");
-                        continue; 
+                        continue;
                     }
 		    if ((pkg_scout->state_status == SS_INSTALLED || pkg_scout->state_want == SW_INSTALL) &&
 		       version_constraints_satisfied(possible_satisfier, pkg_scout) && !is_pkg_a_replaces(pkg_scout,pkg)){
@@ -377,21 +449,22 @@ int version_constraints_satisfied(depend_t * depends, pkg_t * pkg)
     free (temp->version);
     free(temp);
 
-    if((depends->constraint == EARLIER) && 
+    if((depends->constraint == EARLIER) &&
        (comparison < 0))
 	return 1;
-    else if((depends->constraint == LATER) && 
+    else if((depends->constraint == LATER) &&
 	    (comparison > 0))
 	return 1;
-    else if(comparison == 0)
+    else if((depends->constraint == EQUAL) &&
+	    (comparison == 0))
 	return 1;
-    else if((depends->constraint == LATER_EQUAL) && 
+    else if((depends->constraint == LATER_EQUAL) &&
 	    (comparison >= 0))
 	return 1;
-    else if((depends->constraint == EARLIER_EQUAL) && 
+    else if((depends->constraint == EARLIER_EQUAL) &&
 	    (comparison <= 0))
 	return 1;
-    
+
     return 0;
 }
 
@@ -402,7 +475,7 @@ int pkg_dependence_satisfiable(depend_t *depend)
      int n_providers = provider_apkgs->len;
      abstract_pkg_t **apkgs = provider_apkgs->pkgs;
      pkg_vec_t *pkg_vec;
-     int n_pkgs ; 
+     int n_pkgs ;
      int i;
      int j;
 
@@ -449,7 +522,7 @@ int pkg_dependence_satisfied(depend_t *depend)
      return 0;
 }
 
-static int is_pkg_in_pkg_vec(pkg_vec_t * vec, pkg_t * pkg)
+int is_pkg_in_pkg_vec(pkg_vec_t * vec, pkg_t * pkg)
 {
     int i;
     pkg_t ** pkgs = vec->pkgs;
@@ -534,6 +607,65 @@ int pkg_conflicts(pkg_t *pkg, pkg_t *conflictee)
      return 0;
 }
 
+/**
+ * pkg_breaks_reverse_dep returns 1 if pkg does not satisfy the version
+ * constraints of the packages which depend upon pkg and otherwise returns 0.
+ */
+int pkg_breaks_reverse_dep(pkg_t *pkg)
+{
+    /* We consider only the abstract_pkg_t to which pkg belongs (ie. that which
+     * shares its name) not the abstract pkgs which it provides as dependence on
+     * a virtual package should never involve a version constraint.
+     */
+    abstract_pkg_t *rev_dep;
+    abstract_pkg_t *apkg = pkg->parent;
+    unsigned int i = 0;
+    unsigned int j, k, m;
+
+    if (apkg->depended_upon_by == NULL)
+        return 0;
+
+    while ((rev_dep = apkg->depended_upon_by[i++])) {
+        unsigned int npkgs = rev_dep->pkgs->len;
+
+        for (j = 0; j < npkgs; j++) {
+            pkg_t *cmp_pkg = rev_dep->pkgs->pkgs[j];
+            compound_depend_t *cdeps = cmp_pkg->depends;
+            unsigned int ncdeps = cmp_pkg->depends_count;
+
+            /* Only check dependencies of a package which either will be
+             * installed or will remain installed.
+             */
+            if (cmp_pkg->state_want != SW_INSTALL)
+                continue;
+
+            /* Find the dependence on apkg. */
+            for (k = 0; k < ncdeps; k++) {
+                depend_t **deps = cdeps[k].possibilities;
+                unsigned int ndeps = cdeps[k].possibility_count;
+
+                /* Skip recommends and suggests. */
+                if (cdeps[k].type != PREDEPEND && cdeps[k].type != DEPEND)
+                    continue;
+
+                for (m = 0; m < ndeps; m++) {
+                    /* Is this the dependence on apkg? */
+                    if (deps[m]->pkg != apkg)
+                        continue;
+
+                    if (!version_constraints_satisfied(deps[m], pkg)) {
+                        opkg_msg(DEBUG, "Installing %s %s would break reverse dependency from %s.\n",
+                                 pkg->name, pkg->version, cmp_pkg->name);
+                        return 1;
+                    }
+                }
+            }
+        }
+    }
+
+    return 0;
+}
+
 static char ** merge_unresolved(char ** oldstuff, char ** newstuff)
 {
     int oldlen = 0, newlen = 0;
@@ -556,9 +688,9 @@ static char ** merge_unresolved(char ** oldstuff, char ** newstuff)
     return result;
 }
 
-/* 
+/*
  * a kinda kludgy way to back out depends str from two different arrays (reg'l'r 'n pre)
- * this is null terminated, no count is carried around 
+ * this is null terminated, no count is carried around
  */
 char ** add_unresolved_dep(pkg_t * pkg, char ** the_lost, int ref_ndx)
 {
@@ -572,7 +704,7 @@ char ** add_unresolved_dep(pkg_t * pkg, char ** the_lost, int ref_ndx)
     resized = xrealloc(the_lost, sizeof(char *) * (count + 1));
     resized[count - 1] = pkg_depend_str(pkg, ref_ndx);
     resized[count] = NULL;
-    
+
     return resized;
 }
 
@@ -700,7 +832,7 @@ constraint_to_str(enum version_constraint c)
 	case NONE:
 		return "";
 	case EARLIER:
-		return "< ";
+		return "<< ";
 	case EARLIER_EQUAL:
 	       return "<= ";
 	case EQUAL:
@@ -708,12 +840,51 @@ constraint_to_str(enum version_constraint c)
 	case LATER_EQUAL:
 	      return ">= ";
 	case LATER:
-	     return "> ";
+	     return ">> ";
 	}
 
 	return "";
 }
 
+enum version_constraint
+str_to_constraint(char **str)
+{
+	if(!strncmp(*str, "<<", 2)){
+		*str += 2;
+		return EARLIER;
+	}
+	else if(!strncmp(*str, "<=", 2)){
+		*str += 2;
+		return EARLIER_EQUAL;
+	}
+	else if(!strncmp(*str, ">=", 2)){
+		*str += 2;
+		return LATER_EQUAL;
+	}
+	else if(!strncmp(*str, ">>", 2)){
+		*str += 2;
+		return LATER;
+	}
+	else if(!strncmp(*str, "=", 1)){
+		*str += 1;
+		return EQUAL;
+	}
+	/* should these be here to support deprecated designations; dpkg does */
+	else if(!strncmp(*str, "<", 1)){
+		*str += 1;
+		opkg_msg(NOTICE, "Deprecated version constraint '<' was used with the same meaning as '<='. Use '<<' for EARLIER constraint.\n");
+		return EARLIER_EQUAL;
+	}
+	else if(!strncmp(*str, ">", 1)){
+		*str += 1;
+		opkg_msg(NOTICE, "Deprecated version constraint '>' was used with the same meaning as '>='. Use '>>' for LATER constraint.\n");
+		return LATER_EQUAL;
+	}
+	else {
+		return NONE;
+	}
+}
+
 /*
  * Returns a printable string for pkg's dependency at the specified idx. The
  * resultant string must be passed to free() by the caller.
@@ -769,56 +940,62 @@ pkg_depend_str(pkg_t *pkg, int idx)
 	return str;
 }
 
-/*
- * WARNING: This function assumes pre_depends and depends are at the
- * start of the pkg->depends array.
- */
 void buildDependedUponBy(pkg_t * pkg, abstract_pkg_t * ab_pkg)
 {
-     compound_depend_t * depends;
-     int count, othercount;
-     int i, j;
-     abstract_pkg_t * ab_depend;
-     abstract_pkg_t ** temp;
+	compound_depend_t * depends;
+	int count, othercount;
+	int i, j;
+	abstract_pkg_t * ab_depend;
+	abstract_pkg_t ** temp;
 
-     count = pkg->pre_depends_count + pkg->depends_count;
-     depends = pkg->depends;
+	count = pkg->pre_depends_count +
+			pkg->depends_count +
+			pkg->recommends_count +
+			pkg->suggests_count;
 
-     for (i = 0; i < count; i++) {
-	  for (j = 0; j < depends->possibility_count; j++){
-	       ab_depend = depends->possibilities[j]->pkg;
-	       if(!ab_depend->depended_upon_by)
-		    ab_depend->depended_upon_by = xcalloc(1, sizeof(abstract_pkg_t *));
+	for (i = 0; i < count; i++) {
+		depends = &pkg->depends[i];
+		if (depends->type != PREDEPEND
+		    && depends->type != DEPEND
+                    && depends->type != RECOMMEND)
+			continue;
+		for (j = 0; j < depends->possibility_count; j++) {
+			ab_depend = depends->possibilities[j]->pkg;
+			if (!ab_depend->depended_upon_by) {
+				ab_depend->depended_upon_by =
+					xcalloc(1, sizeof(abstract_pkg_t *));
+			}
 
-	       temp = ab_depend->depended_upon_by;
-	       othercount = 1;
-	       while(*temp){
-		    temp++;
-		    othercount++;
-	       }
-	       *temp = ab_pkg;
+			temp = ab_depend->depended_upon_by;
+			othercount = 1;
+			while (*temp) {
+			    temp++;
+			    othercount++;
+			}
+			*temp = ab_pkg;
 
-	       ab_depend->depended_upon_by = xrealloc(ab_depend->depended_upon_by, 
-									(othercount + 1) * sizeof(abstract_pkg_t *));
-	       /* the array may have moved */
-	       temp = ab_depend->depended_upon_by + othercount;
-	       *temp = NULL;
-	  }
-	  depends++;
-     }
+			ab_depend->depended_upon_by =
+				xrealloc(ab_depend->depended_upon_by,
+				(othercount + 1) * sizeof(abstract_pkg_t *));
+
+			/* the array may have been moved by realloc */
+			temp = ab_depend->depended_upon_by + othercount;
+			*temp = NULL;
+		}
+	}
 }
 
 static depend_t * depend_init(void)
 {
-    depend_t * d = xcalloc(1, sizeof(depend_t));    
+    depend_t * d = xcalloc(1, sizeof(depend_t));
     d->constraint = NONE;
     d->version = NULL;
     d->pkg = NULL;
-    
+
     return d;
 }
 
-static int parseDepends(compound_depend_t *compound_depend, 
+static int parseDepends(compound_depend_t *compound_depend,
 			char * depend_str)
 {
      char * pkg_name, buffer[2048];
@@ -852,9 +1029,9 @@ static int parseDepends(compound_depend_t *compound_depend,
 	       *dest++ = *src++;
 	  *dest = '\0';
 	  pkg_name = trim_xstrdup(buffer);
-	
+
 	  /* now look at possible version info */
-	
+
 	  /* skip to next chars */
 	  if(isspace(*src))
 	       while(*src && isspace(*src)) src++;
@@ -862,35 +1039,7 @@ static int parseDepends(compound_depend_t *compound_depend,
 	  /* extract constraint and version */
 	  if(*src == '('){
 	       src++;
-	       if(!strncmp(src, "<<", 2)){
-		    possibilities[i]->constraint = EARLIER;
-		    src += 2;
-	       }
-	       else if(!strncmp(src, "<=", 2)){
-		    possibilities[i]->constraint = EARLIER_EQUAL;
-		    src += 2;
-	       }
-	       else if(!strncmp(src, ">=", 2)){
-		    possibilities[i]->constraint = LATER_EQUAL;
-		    src += 2;
-	       }
-	       else if(!strncmp(src, ">>", 2)){
-		    possibilities[i]->constraint = LATER;
-		    src += 2;
-	       }
-	       else if(!strncmp(src, "=", 1)){
-		    possibilities[i]->constraint = EQUAL;
-		    src++;
-	       }
-	       /* should these be here to support deprecated designations; dpkg does */
-	       else if(!strncmp(src, "<", 1)){
-		    possibilities[i]->constraint = EARLIER_EQUAL;
-		    src++;
-	       }
-	       else if(!strncmp(src, ">", 1)){
-		    possibilities[i]->constraint = LATER_EQUAL;
-		    src++; 
-	       }
+	       possibilities[i]->constraint = str_to_constraint(&src);
 
 	       /* now we have any constraint, pass space to version string */
 	       while(isspace(*src)) src++;
@@ -907,7 +1056,7 @@ static int parseDepends(compound_depend_t *compound_depend,
 	  possibilities[i]->pkg = ensure_abstract_pkg_by_name(pkg_name);
 
 	  free(pkg_name);
-	
+
 	  /* now get past the ) and any possible | chars */
 	  while(*src &&
 		(isspace(*src) ||
