@@ -1,7 +1,7 @@
 /* opkg_hash.c - the opkg package management system
 
    Steven M. Ayer
-   
+
    Copyright (C) 2002 Compaq Computer Corporation
 
    This program is free software; you can redistribute it and/or
@@ -15,13 +15,17 @@
    General Public License for more details.
 */
 
+#include "config.h"
+
 #include <stdio.h>
 
 #include "hash_table.h"
+#include "release.h"
 #include "pkg.h"
 #include "opkg_message.h"
 #include "pkg_vec.h"
 #include "pkg_hash.h"
+#include "parse_util.h"
 #include "pkg_parse.h"
 #include "opkg_utils.h"
 #include "sprintf_alloc.h"
@@ -44,7 +48,7 @@ free_pkgs(const char *key, void *entry, void *data)
 	/* Each entry in the hash table is an abstract package, which contains
 	 * a list of packages that provide the abstract package.
 	 */
-  
+
 	ab_pkg = (abstract_pkg_t*) entry;
 
 	if (ab_pkg->pkgs) {
@@ -69,6 +73,32 @@ pkg_hash_deinit(void)
 	hash_table_deinit(&conf->pkg_hash);
 }
 
+int
+dist_hash_add_from_file(const char *lists_dir, pkg_src_t *dist)
+{
+	nv_pair_list_elt_t *l;
+	char *list_file, *subname;
+
+	list_for_each_entry(l , &conf->arch_list.head, node) {
+		nv_pair_t *nv = (nv_pair_t *)l->data;
+		sprintf_alloc(&subname, "%s-%s", dist->name, nv->name);
+		sprintf_alloc(&list_file, "%s/%s", lists_dir, subname);
+
+		if (file_exists(list_file)) {
+			if (pkg_hash_add_from_file(list_file, dist, NULL, 0)) {
+				free(list_file);
+				return -1;
+			}
+			pkg_src_list_append (&conf->pkg_src_list, subname, dist->value, "__dummy__", 0);
+		}
+
+		free(list_file);
+	}
+
+	return 0;
+}
+
+
 int
 pkg_hash_add_from_file(const char *file_name,
 			pkg_src_t *src, pkg_dest_t *dest, int is_status_file)
@@ -92,8 +122,12 @@ pkg_hash_add_from_file(const char *file_name,
 		pkg->src = src;
 		pkg->dest = dest;
 
-		ret = pkg_parse_from_stream_nomalloc(pkg, fp, 0,
+		ret = parse_from_stream_nomalloc(pkg_parse_line, pkg, fp, 0,
 				&buf, len);
+		if (pkg->name == NULL) {
+			/* probably just a blank line */
+			ret = 1;
+		}
 		if (ret) {
 			pkg_deinit (pkg);
 			free(pkg);
@@ -105,14 +139,22 @@ pkg_hash_add_from_file(const char *file_name,
 			continue;
 		}
 
-		if (!pkg->architecture || !pkg->arch_priority) {
+		if (!pkg->architecture) {
 			char *version_str = pkg_version_str_alloc(pkg);
-			opkg_msg(ERROR, "Package %s version %s has no "
+			opkg_msg(NOTICE, "Package %s version %s has no "
 					"valid architecture, ignoring.\n",
 					pkg->name, version_str);
 			free(version_str);
 			continue;
 		}
+		if (!pkg->arch_priority) {
+			char *version_str = pkg_version_str_alloc(pkg);
+			opkg_msg(DEBUG, "Package %s version %s is built for architecture %s "
+					"which cannot be installed here, ignoring.\n",
+					pkg->name, version_str, pkg->architecture);
+			free(version_str);
+			continue;
+		}
 
 		hash_insert_pkg(pkg, is_status_file);
 
@@ -131,7 +173,7 @@ int
 pkg_hash_load_feeds(void)
 {
 	pkg_src_list_elt_t *iter;
-	pkg_src_t *src;
+	pkg_src_t *src, *subdist;
 	char *list_file, *lists_dir;
 
 	opkg_msg(INFO, "\n");
@@ -139,6 +181,40 @@ pkg_hash_load_feeds(void)
 	lists_dir = conf->restrict_to_default_dest ?
 		conf->default_dest->lists_dir : conf->lists_dir;
 
+	for (iter = void_list_first(&conf->dist_src_list); iter;
+			iter = void_list_next(&conf->dist_src_list, iter)) {
+
+		src = (pkg_src_t *)iter->data;
+
+		sprintf_alloc(&list_file, "%s/%s", lists_dir, src->name);
+
+		if (file_exists(list_file)) {
+			int i;
+			release_t *release = release_new();
+			if(release_init_from_file(release, list_file)) {
+				free(list_file);
+				return -1;
+			}
+
+			unsigned int ncomp;
+			const char **comps = release_comps(release, &ncomp);
+			subdist = (pkg_src_t *) xmalloc(sizeof(pkg_src_t));
+			memcpy(subdist, src, sizeof(pkg_src_t));
+
+			for(i = 0; i < ncomp; i++){
+				subdist->name = NULL;
+				sprintf_alloc(&subdist->name, "%s-%s", src->name, comps[i]);
+				if (dist_hash_add_from_file(lists_dir, subdist)) {
+					free(subdist->name); free(subdist);
+					free(list_file);
+					return -1;
+				}
+			}
+			free(subdist->name); free(subdist);
+		}
+		free(list_file);
+	}
+
 	for (iter = void_list_first(&conf->pkg_src_list); iter;
 			iter = void_list_next(&conf->pkg_src_list, iter)) {
 
@@ -171,7 +247,7 @@ pkg_hash_load_status_files(void)
 
 	for (iter = void_list_first(&conf->pkg_dest_list); iter;
 			iter = void_list_next(&conf->pkg_dest_list, iter)) {
-	
+
 		dest = (pkg_dest_t *)iter->data;
 
 		if (file_exists(dest->status_file_name)) {
@@ -197,7 +273,6 @@ pkg_hash_fetch_best_installation_candidate(abstract_pkg_t *apkg,
      int i, j;
      int nprovides = 0;
      int nmatching = 0;
-     int wrong_arch_found = 0;
      pkg_vec_t *matching_pkgs;
      abstract_pkg_vec_t *matching_apkgs;
      abstract_pkg_vec_t *provided_apkg_vec;
@@ -227,6 +302,11 @@ pkg_hash_fetch_best_installation_candidate(abstract_pkg_t *apkg,
      /* accumulate all the providers */
      for (i = 0; i < nprovides; i++) {
 	  abstract_pkg_t *provider_apkg = provided_apkgs[i];
+
+          /* Don't double insert packages. */
+          if (abstract_pkg_vec_contains(providers, provider_apkg))
+              continue;
+
 	  opkg_msg(DEBUG, "Adding %s to providers.\n", provider_apkg->name);
 	  abstract_pkg_vec_insert(providers, provider_apkg);
      }
@@ -247,7 +327,7 @@ pkg_hash_fetch_best_installation_candidate(abstract_pkg_t *apkg,
 	  }
 
 	  if (replacement_apkg)
-	       opkg_msg(DEBUG, "replacement_apkg=%s for provider_apkg=%s.\n", 
+	       opkg_msg(DEBUG, "replacement_apkg=%s for provider_apkg=%s.\n",
 			    replacement_apkg->name, provider_apkg->name);
 
 	  if (replacement_apkg && (replacement_apkg != provider_apkg)) {
@@ -262,37 +342,60 @@ pkg_hash_fetch_best_installation_candidate(abstract_pkg_t *apkg,
 			       provider_apkg->name);
 	       continue;
 	  }
-    
 
-	  /* now check for supported architecture */
-	  {
-	       int max_count = 0;
 
-	       /* count packages matching max arch priority and keep track of last one */
-	       for (j=0; j<vec->len; j++) {
-		    pkg_t *maybe = vec->pkgs[j];
-		    opkg_msg(DEBUG, "%s arch=%s arch_priority=%d version=%s.\n",
-				 maybe->name, maybe->architecture,
-				 maybe->arch_priority, maybe->version);
-                    /* We make sure not to add the same package twice. Need to search for the reason why 
-                       they show up twice sometimes. */
-		    if ((maybe->arch_priority > 0) && (! pkg_vec_contains(matching_pkgs, maybe))) {
-			 max_count++;
-			 abstract_pkg_vec_insert(matching_apkgs, maybe->parent);
-			 pkg_vec_insert(matching_pkgs, maybe);
-		    }
-	       }
+	  for (j=0; j<vec->len; j++) {
+	      pkg_t *maybe = vec->pkgs[j];
 
-		if (vec->len > 0 && matching_pkgs->len < 1)
-			wrong_arch_found = 1;
+              /* If package is installed and held, add it to the matching list
+               * regardless of whether it satisfies our other checks. This
+               * ensures that a held package won't be removed due to edge cases
+               * like the forced installation of a new package with dependency
+               * on a later version of the held package.
+               */
+              if ((maybe->state_status == SS_INSTALLED || maybe->state_status == SS_UNPACKED)
+                      && maybe->state_flag == SF_HOLD)
+                  goto add_matching_pkg;
+
+              /* Ensure that the package meets the specified constraint. */
+              if (constraint_fcn && !constraint_fcn(maybe, cdata)) {
+                  opkg_msg(DEBUG, "Not selecting %s %s due to unmatched constraint.\n",
+                           maybe->name, maybe->version);
+                  continue;
+              }
+
+	      /* Ensure that installing this package won't break the
+	       * dependencies of packages which are already installed, unless
+	       * force_depends is set.
+	       */
+	      if (pkg_breaks_reverse_dep(maybe) &&
+		      !conf->force_depends) {
+		  opkg_msg(NOTICE, "Not selecting %s %s as installing it would break existing dependencies.\n",
+			   maybe->name, maybe->version);
+		  continue;
+	      }
+
+	      /* now check for supported architecture */
+	      opkg_msg(DEBUG, "%s arch=%s arch_priority=%d version=%s.\n",
+			      maybe->name, maybe->architecture,
+			      maybe->arch_priority, maybe->version);
+              if (maybe->arch_priority <= 0) {
+                  opkg_msg(NOTICE, "Not selecting %s %s due to incompatible architecture.\n",
+                           maybe->name, maybe->version);
+                  continue;
+              }
+
+add_matching_pkg:
+	      /* We make sure not to add the same package twice. Need to search for the reason why
+		  they show up twice sometimes. */
+	      if (! pkg_vec_contains(matching_pkgs, maybe)) {
+	          abstract_pkg_vec_insert(matching_apkgs, maybe->parent);
+		  pkg_vec_insert(matching_pkgs, maybe);
+	      }
 	  }
      }
 
      if (matching_pkgs->len < 1) {
-	  if (wrong_arch_found)
-	        opkg_msg(ERROR, "Packages for %s found, but"
-			" incompatible with the architectures configured\n",
-			apkg->name);
           pkg_vec_free(matching_pkgs);
           abstract_pkg_vec_free(matching_apkgs);
           abstract_pkg_vec_free(providers);
@@ -307,13 +410,28 @@ pkg_hash_fetch_best_installation_candidate(abstract_pkg_t *apkg,
 
      for (i = 0; i < matching_pkgs->len; i++) {
 	  pkg_t *matching = matching_pkgs->pkgs[i];
-          if (constraint_fcn(matching, cdata)) {
-             opkg_msg(DEBUG, "Candidate: %s %s.\n",
-			     matching->name, matching->version) ;
-             good_pkg_by_name = matching;
-	     /* It has been provided by hand, so it is what user want */
-             if (matching->provided_by_hand == 1)
-                break;                                 
+          /* Set good_pkg_by_name if the package name matches the originally
+           * requested name (which will be the apkg name).
+           */
+          if (strcmp(matching->name, apkg->name) == 0) {
+              opkg_msg(DEBUG, "Candidate: %s %s.\n",
+                      matching->name, matching->version) ;
+              /* It has been provided by hand, so it is what user want */
+              if (matching->provided_by_hand == 1) {
+                  good_pkg_by_name = matching;
+                  break;
+              }
+              /* Respect to the arch priorities when given alternatives */
+              if (good_pkg_by_name && conf->prefer_arch_to_version) {
+                  if (matching->arch_priority >= good_pkg_by_name->arch_priority) {
+                      good_pkg_by_name = matching;
+                      opkg_msg(DEBUG, "%s %s wins by priority.\n",
+                          matching->name, matching->version) ;
+                  } else
+                      opkg_msg(DEBUG, "%s %s wins by priority.\n",
+                          good_pkg_by_name->name, good_pkg_by_name->version) ;
+              } else
+                  good_pkg_by_name = matching;
           }
      }
 
@@ -344,7 +462,7 @@ pkg_hash_fetch_best_installation_candidate(abstract_pkg_t *apkg,
 				matching->name, prio);
                   }
               }
-          
+
           }
 
      if (conf->verbosity >= INFO && matching_apkgs->len > 1) {
@@ -364,7 +482,7 @@ pkg_hash_fetch_best_installation_candidate(abstract_pkg_t *apkg,
      abstract_pkg_vec_free(matching_apkgs);
      abstract_pkg_vec_free(providers);
 
-     if (good_pkg_by_name) {   /* We found a good candidate, we will install it */ 
+     if (good_pkg_by_name) {   /* We found a good candidate, we will install it */
 	  return good_pkg_by_name;
      }
      if (held_pkg) {
@@ -396,17 +514,6 @@ pkg_hash_fetch_best_installation_candidate(abstract_pkg_t *apkg,
      return NULL;
 }
 
-static int
-pkg_name_constraint_fcn(pkg_t *pkg, void *cdata)
-{
-	const char *name = (const char *)cdata;
-
-	if (strcmp(pkg->name, name) == 0)
-		return 1;
-	else
-		return 0;   
-}
-
 static pkg_vec_t *
 pkg_vec_fetch_by_name(const char *pkg_name)
 {
@@ -439,7 +546,7 @@ pkg_hash_fetch_best_installation_candidate_by_name(const char *name)
 		return NULL;
 
 	return pkg_hash_fetch_best_installation_candidate(apkg,
-				pkg_name_constraint_fcn, apkg->name, 0);
+                NULL, NULL, 0);
 }
 
 
@@ -449,10 +556,10 @@ pkg_hash_fetch_by_name_version(const char *pkg_name, const char * version)
 	pkg_vec_t * vec;
 	int i;
 	char *version_str = NULL;
-    
+
 	if(!(vec = pkg_vec_fetch_by_name(pkg_name)))
 		return NULL;
-    
+
 	for(i = 0; i < vec->len; i++) {
 		version_str = pkg_version_str_alloc(vec->pkgs[i]);
 		if(!strcmp(version_str, version)) {
@@ -464,7 +571,7 @@ pkg_hash_fetch_by_name_version(const char *pkg_name, const char * version)
 
 	if(i == vec->len)
 		return NULL;
-    
+
 	return vec->pkgs[i];
 }
 
@@ -619,30 +726,43 @@ hash_insert_pkg(pkg_t *pkg, int set_status)
 	pkg->parent = ab_pkg;
 }
 
+static const char *
+strip_offline_root(const char *file_name)
+{
+	unsigned int len;
+
+	if (conf->offline_root) {
+		len = strlen(conf->offline_root);
+		if (strncmp(file_name, conf->offline_root, len) == 0)
+			file_name += len;
+	}
+
+	return file_name;
+}
+
+void
+file_hash_remove(const char *file_name)
+{
+	file_name = strip_offline_root(file_name);
+	hash_table_remove(&conf->file_hash, file_name);
+}
 
 pkg_t *
 file_hash_get_file_owner(const char *file_name)
 {
-	return hash_table_get(&conf->file_hash, file_name); 
+	file_name = strip_offline_root(file_name);
+	return hash_table_get(&conf->file_hash, file_name);
 }
 
 void
 file_hash_set_file_owner(const char *file_name, pkg_t *owning_pkg)
 {
-	pkg_t *old_owning_pkg = hash_table_get(&conf->file_hash, file_name);
-	int file_name_len = strlen(file_name);
+	pkg_t *old_owning_pkg;
 
-	if (file_name[file_name_len -1] == '/')
-		return;
+	file_name = strip_offline_root(file_name);
 
-	if (conf->offline_root) {
-		unsigned int len = strlen(conf->offline_root);
-		if (strncmp(file_name, conf->offline_root, len) == 0) {
-			file_name += len;
-		}
-	}
-
-	hash_table_insert(&conf->file_hash, file_name, owning_pkg); 
+	old_owning_pkg = hash_table_get(&conf->file_hash, file_name);
+	hash_table_insert(&conf->file_hash, file_name, owning_pkg);
 
 	if (old_owning_pkg) {
 		pkg_get_installed_files(old_owning_pkg);
