@@ -15,7 +15,7 @@
    General Public License for more details.
 */
 
-#include "includes.h"
+#include "config.h"
 
 #include "pkg_src.h"
 #include "libbb/libbb.h"
